#include <stdio.h>
#include <stdlib.h>
#include "trace.h"

FILE* tf = NULL;
int ref_count = 0;
int read_count = 0;
int write_count = 0;

void parse_args( CacheInfo* ci, int argc, char** argv )
{
    int i;
    if ( argc != 6 )
    {
	fprintf( stderr, "usage: %s cachesize organization blocksize policy tracefile\n", argv[0] );
	fprintf( stderr, "   organization = dm | 2way\n" );
	fprintf( stderr, "   policy = wt | wb\n" );
	exit( 1 );
    }
    
    ci->cachesize = atoi( argv[1] );
    for ( i = 16; i <= 1024; i <<= 1 )
	if ( ci->cachesize == i )
	    break;
    if ( i > 1024 )
    {
	fprintf( stderr, "error: cachesize must be a power of 2 from 16 to 1024\n" );
	exit( 1 );
    }

    if ( strcmp( argv[2], "dm" ) == 0 )
	ci->organization = DirectMapped;
    else if ( strcmp( argv[2], "2way" ) == 0 )
	ci->organization = TwoWayLRU;
    else
    {
	fprintf( stderr, "error: organization must be DM or 2WAY\n" );
	exit( 1 );
    }

    ci->blocksize = atoi( argv[3] );
    for ( i = 8; i <= ci->cachesize >> 1; i <<= 1 )
	if ( ci->blocksize == i )
	    break;
    if ( i > ci->cachesize >> 1 )
    {
	fprintf( stderr, "error: blocksize must be a power of 2 from 8 to cachesize/2\n" );
	exit( 1 );
    }

    if ( strcmp( argv[4], "wt" ) == 0 )
	ci->policy = WriteThrough;
    else if ( strcmp( argv[4], "wb" ) == 0 )
	ci->policy = WriteBack;
    else
    {
	fprintf( stderr, "error: organization must be WT or WB\n" );
	exit( 1 );
    }

    switch ( start_trace_file( argv[5] ) )
    {
      case TraceOk:
	break;

      case TraceFileOpenFailed:
	fprintf( stderr, "error: open of trace file failed\n" );
	exit( 1 );
	break;

      default:
	fprintf( stderr, "error: start_trace_file failed mysteriously\n" );
	exit( 1 );
    }

    printf( "  cache size:  %d bytes\n", ci->cachesize );
    printf( "organization:  %s\n", ci->organization == DirectMapped ? "direct mapped" : "2-way, LRU" );
    printf( "  block size:  %d bytes\n", ci->blocksize );
    printf( "      policy:  %s\n", ci->policy == WriteThrough ? "write-through" : "write-back" );
    printf( "  trace file:  %s\n\n", argv[5] );
}    

TraceCode start_trace_file( char* fn )
{
    if ( tf )
    {
	fclose( tf );
	tf = NULL;
    }

    tf = fopen( fn, "r" );
    if ( tf == NULL )
    {
	return TraceFileOpenFailed;
    }

    return TraceOk;
}

TraceCode get_next_reference( unsigned int* address, int* count )
{
    int type;

    if ( tf == NULL )
	return TraceNoTraceFile;
    
    if ( fscanf( tf, "%d %x\n", &type, address ) != 2 )
    {
	fclose( tf );
	tf = NULL;
	return TraceBadTraceFile;
    }

    if ( type == 0 )
    {
	fclose( tf );
	tf = NULL;
	return TraceEOF;
    }

    if ( count )
	*count = ref_count;
    ++ref_count;

    if ( type == TraceRead )
	++read_count;
    else
	++write_count;
    
    return (TraceCode)type;
}

void show_statistics( int rh, int rm, int wh, int wm, int br, int bw )
{
    printf( "\n\ncache simulation results\n------------------------\n" );
    printf( "trace contained %d reads and %d writes\n", read_count, write_count );

    if ( rh+rm )
	printf( " reads:  %4d hits  %4d misses  (%5d total)  hit ratio = %6.3f\n", rh, rm, rh+rm, (float)rh/(rh+rm) );
    else
	printf( " reads:  %4d hits  %4d misses  (%5d total)  hit ratio =  -.-\n", rh, rm, rh+rm );

    if ( wh+wm )
	printf( "writes:  %4d hits  %4d misses  (%5d total)  hit ratio = %6.3f\n", wh, wm, wh+wm, (float)wh/(wh+wm) );
    else
	printf( "writes:  %4d hits  %4d misses  (%5d total)  hit ratio =  -.-\n", wh, wm, wh+wm );

    if ( read_count )
	printf( "         %6d bytes read from memory   (%6.1f%%)\n", br, (float)br*100.0/(4*read_count) );
    else
	printf( "         %6d bytes read from memory   (   -.-%%)\n", br );
    
    if ( write_count )
	printf( "         %6d bytes written to memory  (%6.1f%%)\n", bw, (float)bw*100.0/(4*write_count) );
    else
	printf( "         %6d bytes written to memory  (   -.-%%)\n", bw );
}
    
    
	    

